import os
import mmap
import struct
import numpy
import copy

from trsfile.trace import Trace
from trsfile.common import Header, SampleCoding, TracePadding
from trsfile.engine.engine import Engine

class TrsEngine(Engine):
	"""
	This engine supports .trs files from Riscure as specified in the
	"Trace set coding" document in Inspector.

	This engine supports the following options:

	+--------------+-----------------------------------------------------------+
	| Option       | Description                                               |
	+==============+===========================================================+
	| headers      | Dictionary containing zero or more headers, see           |
	|              | :py:class:`trsfile.common.Header`                         |
	+--------------+-----------------------------------------------------------+
	| live_update  | Performs live update of the TRS file every N traces. True |
	|              | for updating after every trace and False for never.       |
	+--------------+-----------------------------------------------------------+
	| padding_mode | Padding mode to use. The supported values are:            |
	|              | :py:attr:`trsfile.common.TracePadding.NONE` (default)     |
	|              | :py:attr:`trsfile.common.TracePadding.AUTO`               |
	+--------------+-----------------------------------------------------------+
	"""

	def __init__(self, path, mode = 'x', **options):
		self.path = path
		self.handle = None
		self.file_handle = None

		self.data_offset = None
		self.sample_length = None
		self.trace_length = None

		self.is_mmap_synched = False

		# Initialize empty dictionaries
		self.headers = {}
		self.header_locations = {}

		# Get the options
		headers = options.get('headers', None)
		self.live_update = int(options.get('live_update', False))
		self.live_update_count = 0
		self.padding_mode = options.get('padding_mode', TracePadding.NONE)
		if not isinstance(self.padding_mode, TracePadding):
			raise TypeError('TrsFile requires padding_mode to be of type \'TracePadding\'')
		if self.padding_mode not in [TracePadding.NONE, TracePadding.AUTO]:
			raise ValueError('TrsFile only supports the padding mode NONE and AUTO')

		# Parse the mode
		if mode == 'r':
			"""r = open for reading"""
			if headers is not None:
				raise TypeError('Cannot change headers when reading TRS files.')

			if not os.path.isfile(self.path):
				raise FileNotFoundError('No TRS file: \'{0:s}\''.format(path))

			self.file_handle = open(path, 'rb')
			self.handle = mmap.mmap(self.file_handle.fileno(), 0, access = mmap.ACCESS_READ)
			self.read_only = True
			self.read_headers = True

		elif mode == 'w':
			"""open for writing, truncating the file first"""
			if headers is not None and any(not isinstance(header, Header) for header in headers):
				raise TypeError('Creation of TRS files requires passing Headers to the constructor.')

			# Sadly, to memory map we need a file with a minimum of length 1
			self.file_handle = open(path, 'wb')
			self.file_handle.write(b'\x00')
			self.file_handle.close()

			# Now we can open it properly
			self.file_handle = open(path, 'r+b')
			self.handle = mmap.mmap(self.file_handle.fileno(), 0, access = mmap.ACCESS_WRITE)

			self.read_only = False
			self.read_headers = False

		elif mode == 'x':
			"""open for exclusive creation, failing if the file already exists"""
			if headers is not None and any(not isinstance(header, Header) for header in headers):
				raise TypeError('Creation of TRS files requires passing Headers to the constructor.')

			if os.path.isfile(self.path):
				raise FileExistsError('TRS file exists: \'{0:s}\''.format(self.path))

			# Sadly, to memory map we need a file with a minimum of length 1
			self.file_handle = open(path, 'wb')
			self.file_handle.write(b'\x00')
			self.file_handle.close()

			self.file_handle = open(path, 'r+b')
			self.handle = mmap.mmap(self.file_handle.fileno(), 0, access = mmap.ACCESS_WRITE)
			self.read_only = False
			self.read_headers = False

		elif mode == 'a':
			"""a = open for writing, appending to the end of the file if it exists"""

			# Check if the file exists, if so, we read in the headers and that will be our LAW!
			if os.path.isfile(self.path):
				self.read_headers = True
			else:
				self.read_headers = False

				# We need to create an empty file
				# Sadly, to memory map we need a file with a minimum of length 1
				self.file_handle = open(path, 'wb')
				self.file_handle.write(b'\x00')
				self.file_handle.close()

			if self.read_headers and headers is not None:
				raise TypeError('Cannot change headers when reading TRS files.')
			elif not self.read_headers and headers is not None and any(not isinstance(header, Header) for header in headers):
				raise TypeError('Creation of TRS files requires passing instances of Headers to the constructor.')

			# NOTE: We are using r+b mode because we are essentially updating the file!
			self.file_handle = open(path, 'r+b')
			self.handle = mmap.mmap(self.file_handle.fileno(), 0, access = mmap.ACCESS_WRITE)
			self.read_only = False

		else:
			raise ValueError('invalid mode: \'{0:s}\''.format(mode))

		self.__initialize_headers(headers)

	def length(self):
		return self.headers.get(Header.NUMBER_TRACES, 0)

	def is_closed(self):
		return self.handle is None or self.handle.closed

	def set_traces(self, index, traces):
		# Make sure we have proper indexing
		if isinstance(index, slice):
			start, stop, step = index.indices(index.stop)
			indexes = range(start, max(stop, start + len(traces)), step)
		else:
			indexes = range(index, index + 1)

		# Check if we are appending traces directly to the end of the file
		if indexes.start > self.headers[Header.NUMBER_TRACES]:
			raise IndexError('No arbitrary indexing supported')

		# Make sure we have enough traces for every index
		if len(indexes) != len(traces):
			raise TypeError('The number of provided traces ({0:d}) have to be equal to the number of indexes ({1:d})'.format(len(traces), len(indexes)))

		# Early return
		if len(indexes) <= 0:
			return

		# Check if any of the following headers are NOT initialized:
		# - NUMBER_SAMPLES
		# - LENGTH_DATA
		# - TITLE_SPACE
		# For any of these uninitialized headers, check if all traces are the same
		# for these fields, and set them to that value.
		if self.padding_mode == TracePadding.AUTO:
			update_headers = {}
			if self.headers[Header.NUMBER_SAMPLES] is None:
				lengths = set([len(trace) for trace in traces])

				# Check padding mode on how we are going to do this
				update_headers[Header.NUMBER_SAMPLES] = max(lengths)

			if self.headers[Header.LENGTH_DATA] is None:
				if len(set([len(trace.data) for trace in traces])) > 1:
					raise TypeError('Traces have different data length, this is not supported in TRS files')
				update_headers[Header.LENGTH_DATA] = len(traces[0].data)

			if self.headers[Header.SAMPLE_CODING] is None:
				if len(set([trace.sample_coding for trace in traces])) > 1:
					raise TypeError('Traces have different sample coding, this is not supported in TRS files')
				update_headers[Header.SAMPLE_CODING] = traces[0].sample_coding

			if self.headers[Header.TITLE_SPACE] is None:
				update_headers[Header.TITLE_SPACE] = max([len(trace.title) for trace in traces])

			# Now update headers
			self.update_headers(update_headers)
		elif self.padding_mode == TracePadding.NONE:
			# We need to verify if all required headers are set, else throw an error
			required_headers = [Header.NUMBER_SAMPLES, Header.LENGTH_DATA, Header.SAMPLE_CODING, Header.TITLE_SPACE]
			missing_headers = [header for header in required_headers if header not in self.headers or self.headers[header] is None]
			if len(missing_headers) > 0:
				raise ValueError('The following headers are not set: ' + ', '.join([h.name for h in missing_headers]) + ', set these headers or use PaddingMode.AUTO')

		else:
			raise NotImplementedError('This padding mode is not supported')

		# Pre-compute some static information based on headers
		if self.sample_length is None:
			self.sample_length = self.headers[Header.NUMBER_SAMPLES] * self.headers[Header.SAMPLE_CODING].size
		if self.trace_length is None:
			self.trace_length = self.sample_length + self.headers.get(Header.LENGTH_DATA, 0) + self.headers.get(Header.TITLE_SPACE, 0)

		# Store all traces with the next sequence numbers and keep these numbers as a list
		for i, trace in zip(indexes, traces):
			# Update the trace headers to be a reference to our internal headers because that is how it is!
			trace.headers = self.headers

			# Check padding mode
			if self.padding_mode == TracePadding.NONE and len(trace) != self.headers[Header.NUMBER_SAMPLES]:
				raise ValueError('Trace has a different length from the expected length and padding mode is NONE')

			# Seek to the beginning of the trace (this automatically enables us to overwrite)
			self.file_handle.seek(self.data_offset + i * self.trace_length)

			# Title and title padding
			title = trace.title.strip().encode('utf-8')
			if len(title) > self.headers[Header.TITLE_SPACE]:
				raise TypeError('Trace title is longer than available title space')
			self.file_handle.write(title)
			if len(title) < self.headers[Header.TITLE_SPACE]:
				self.file_handle.write(bytes(self.headers[Header.TITLE_SPACE] - len(title)))

			# Data and data padding
			if self.padding_mode == TracePadding.NONE and len(trace.data) > self.headers[Header.LENGTH_DATA]:
				raise TypeError('Trace data is longer than available data space')
			self.file_handle.write(trace.data[0:self.headers[Header.LENGTH_DATA]])
			if len(trace.data) < self.headers[Header.LENGTH_DATA]:
				self.file_handle.write(bytes(self.headers[Header.LENGTH_DATA] - len(trace.data)))

			# Automatic truncate
			trace.samples[:self.headers[Header.NUMBER_SAMPLES]].tofile(self.file_handle)

			# Add any required padding
			length = (self.headers[Header.NUMBER_SAMPLES] - len(trace.samples)) * self.headers[Header.SAMPLE_CODING].size
			if length > 0:
				self.file_handle.write(length * b'\x00')

		# Write the new total number of traces
		# If you want to have live update, you can give this flag and have this
		# automatically write to the file
		new_number_traces = max(self.headers[Header.NUMBER_TRACES], max(indexes) + 1)
		if self.headers[Header.NUMBER_TRACES] < new_number_traces:
			self.is_mmap_synched = False
			self.live_update_count += len(traces)

			if self.live_update != 0 and self.live_update_count >= self.live_update:
				self.live_update_count = 0
				self.update_header(Header.NUMBER_TRACES, new_number_traces)

				# Force flush
				self.handle.flush()
				self.file_handle.flush()
				os.fsync(self.file_handle.fileno())
			else:
				self.headers[Header.NUMBER_TRACES] = new_number_traces

	def get_traces(self, index):
		# check for slicing
		if isinstance(index, slice):
			# No bounds checking when using slices, that's how python rolls!
			indexes = range(*index.indices(self.length()))
		else:
			# Wrap around for negative index
			if index < 0:
				index = index % self.length()

			# Check if we are still in bounds
			if index >= self.length():
				raise IndexError('List index out of range')

			indexes = range(index, index + 1)

		# We need to resize the mmap if we added something directly on the file handle
		# We do it here for optimization purposes, if you do not read, no resizing :)
		if not self.is_mmap_synched and not self.read_only:
			total_file_size = self.data_offset + (self.length() + 1) * self.trace_length
			if self.handle.size() < total_file_size:
				self.handle.resize(total_file_size)
			self.is_mmap_synched = True

		# Now read in all traces
		traces = []
		for i in indexes:
			# Seek to the beginning of the trace
			self.handle.seek(self.data_offset + i * self.trace_length)

			# Read the title
			if Header.TITLE_SPACE in self.headers:
				title = self.handle.read(self.headers[Header.TITLE_SPACE]).rstrip(b'\x00').decode('utf-8')
			else:
				title = ''     # No title

			# Read data
			if Header.LENGTH_DATA in self.headers:
				data = self.handle.read(self.headers[Header.LENGTH_DATA])
			else:
				data = bytes() # No data

			# Read all the samples
			samples = numpy.frombuffer(self.handle.read(self.trace_length), self.headers[Header.SAMPLE_CODING].format, self.headers[Header.NUMBER_SAMPLES])

			traces.append(Trace(self.headers[Header.SAMPLE_CODING], samples, data, title, self.headers))

		return traces

	def close(self):
		"""Closes the open file handle if it is opened"""

		# Close all handles
		if self.handle is not None and not self.handle.closed:
			# Make sure we write all headers to the file
			if not self.read_only:
				self.__write_headers({Header.NUMBER_TRACES: self.headers[Header.NUMBER_TRACES]})

			# Flush the mmap (according to docs this is important) and close
			self.handle.flush()
			self.handle.close()
		if self.file_handle is not None and not self.file_handle.closed:
			self.file_handle.close()

	def update_headers(self, headers):
		changed_headers = super().update_headers(headers)
		if len(changed_headers) > 0:
			self.__write_headers(changed_headers)

	def __initialize_headers(self, headers = None):
		"""Initialize the headers, this is done either by reading the headers from file or using headers given on creation"""
		if self.read_headers:
			self.__read_headers()
		else:
			self.__create_headers(headers)

	def __create_headers(self, headers):
		if headers is not None:
			self.headers = copy.deepcopy(headers)

		# Let's support dynamic sample coding depending on the trace
		if Header.SAMPLE_CODING not in self.headers:
			self.headers[Header.SAMPLE_CODING] = None

		# Add any mandatory headers that are missing
		for header in Header.get_mandatory():
			if not header in self.headers:
				self.headers[header] = header.default

		# Finally add some extra headers that are freaking useful if they are not provided
		# This is up for debate if somethings are missing
		if Header.TITLE_SPACE not in self.headers:
			self.headers[Header.TITLE_SPACE] = Header.TITLE_SPACE.default
		if Header.LENGTH_DATA not in self.headers:
			self.headers[Header.LENGTH_DATA] = None

		# Write the initial headers
		self.__write_headers()

	def __write_headers(self, headers = None):
		if headers is None:
			headers = self.headers

		# Check if we have any work to do
		if len(headers) <= 0:
			return

		# Save the headers
		for header, value in headers.items():
			# Skip TRACE_BLOCK header as we write that last!
			if header == Header.TRACE_BLOCK:
				continue

			# Check if we have definitions for all headers
			if not isinstance(header, Header):
				raise TypeError('Cannot write unknown header to trace set')

			# Obtain the tag value
			if header.type is int:
				tag_value = b'\xff' * header.length if value is None else value.to_bytes(header.length, 'little')
			elif header.type is float:
				tag_value = struct.pack('<f', 0.0 if value is None else value)
			elif header.type is bool:
				tag_value = struct.pack('<?', 0 if value is None else value)
			elif header.type is str:
				tag_value = value.encode('utf-8')
			elif header.type is SampleCoding:
				tag_value = b'\xff' if value is None else value.value.to_bytes(1, 'little')
			elif header.type is bytes:
				tag_value = value
			else:
				raise TypeError('Header has a type that can not be serialized')

			# The tag length is easy!
			tag_length = len(tag_value)

			# Store or reuse the header_locations
			if header in self.header_locations:
				if self.header_locations[header][1] != tag_length:
					raise TypeError('While updating a header, the length of the value changed which is not supported')

				# Update the TLV value
				offset = self.header_locations[header][0]
				self.handle[offset : offset + len(tag_value)] = tag_value
			else:
				# Construct the TLV
				tag = [header.value]
				if tag_length >= 0x80:
					tag_length_length = math.ceil(tag_length.bit_length() / 8.0)
					tag += [0x80 | tag_length_length] + tag_length.to_bytes(tag_length_length, 'little')
				else:
					tag += [tag_length]
				tag += tag_value

				# Store this index for future references
				if self.handle.size() < self.handle.tell() + len(tag):
					self.handle.resize(self.handle.tell() + len(tag))
				self.handle.write(bytes(tag))
				self.header_locations[header] = (self.handle.tell() - len(tag_value), tag_length)

		# Save the TRACE_BLOCK if not already saved
		if Header.TRACE_BLOCK not in self.header_locations:
			# Write the TRACE_BLOCK
			if self.handle.size() < self.handle.tell() + 2:
				self.handle.resize(self.handle.tell() + 2)
			self.handle.write(bytes([Header.TRACE_BLOCK.value, 0]))

			# Calculate offset
			self.data_offset = self.handle.tell()
			self.header_locations[Header.TRACE_BLOCK] = None
		elif self.data_offset is None:
			# This should never happen, but who knows?!
			raise NotImplementedError('Data offset is still None but TRACE_BLOCK TLV already in headers?!?!?!')

	def __read_headers(self):
		"""Read all internal headers from the file"""
		self.headers = {}
		self.header_locations = {}

		# Jump to the beginning of the file (should contain TLV)
		self.handle.seek(0)

		# Parse all headers until the TRACE_BLOCK
		while Header.TRACE_BLOCK not in self.headers:
			# Obtain the Tag
			tag = self.handle.read(1)[0]

			# Obtain the Length
			tag_length = self.handle.read(1)[0]

			if (tag_length & 0x80) != 0:
				tag_length = int.from_bytes(self.handle.read(tag_length & 0x7F), 'little')

			# Obtain the Value
			tag_value_index = self.handle.tell()
			tag_value = self.handle.read(tag_length) if tag_length > 0 else None

			# Interpreter it
			header = None
			if Header.has_value(tag):
				header = Header(tag)
				if header.type is int:
					tag_value = int.from_bytes(tag_value, 'little')
				elif header.type is float:
					tag_value, = struct.unpack('<f', tag_value)
				elif header.type is bool:
					tag_value, = struct.unpack('<?', tag_value)
				elif header.type is str:
					tag_value = tag_value.decode('utf-8')
				elif header.type is SampleCoding:
					tag_value = SampleCoding(tag_value[0])
			else:
				raise NotImplementedError('Warning: tag {tag:02X} is not supported by the library, please submit an issue on Github.'.format(tag=tag))

			self.headers[tag if header is None else header] = tag_value
			self.header_locations[tag if header is None else header] = (tag_value_index, tag_length)

		# Sanity: Check if we have all mandatory headers, if not, throw an error if we are in reading mode!
		if not Header.get_mandatory().issubset(self.headers):
			raise IOError('TRS file does not contain all mandatory headers')

		# Pre-compute some static information based on headers
		self.data_offset = self.handle.tell()
		self.sample_length = self.headers[Header.NUMBER_SAMPLES] * self.headers[Header.SAMPLE_CODING].size
		self.trace_length = self.sample_length + self.headers.get(Header.LENGTH_DATA, 0) + self.headers.get(Header.TITLE_SPACE, 0)

		# Sanity: Check if the file has the proper size
		self.handle.seek(0, os.SEEK_END)
		file_size = self.handle.tell()
		if file_size != self.data_offset + self.headers[Header.NUMBER_TRACES] * self.trace_length:
			raise IOError('TRS file has an unexpected length')



import os
import mmap
import struct
import numpy
import copy

from trsfile.trace import Trace
from trsfile.common import Header, SampleCoding, TracePadding
from trsfile.engine.engine import Engine

# All our engines
from trsfile.engine.trs import TrsEngine
from trsfile.engine.file import FileEngine

engines = {
	'trsengine': TrsEngine,
	'fileengine': FileEngine,
}
ENGINE = 'engine'

class TraceSet:
	"""The :py:obj:`TraceSet` class behaves like a :py:obj:`list`
	object were each item in the list is a :py:obj:`Trace`.

	Storing the :py:obj:`TraceSet` requires knowledge on the format which is
	resolved through the usage of storage engines (:py:obj:`Engine`).
	"""

	def __init__(self, path, mode = 'r', **options):
		# Defaults
		self.engine = None

		# Get the storage engine if one is given, else default to TrsEngine
		engine = options.get(ENGINE, TrsEngine)
		if ENGINE in options:
			del options[ENGINE]

		# We also support engine to be passed as string
		if isinstance(engine, str):
			engine_name = engine.lower()
			if not engine_name.endswith(ENGINE):
				engine_name += ENGINE
			if engine_name not in engines:
				raise ValueError('The storage engine \'{0:s}\'does not exists'.format(engine_name))
			engine = engines[engine_name]

		# Check type
		if not issubclass(engine, Engine):
			raise TypeError('The storage engine has to be of type \'Engine\'')

		self.engine = engine(path, mode, **options)

	def __del__(self):
		self.close()

	def __iter__(self):
		""" reset pointer """
		self.iterator_index = -1
		return self

	def __next__(self):
		self.iterator_index = self.iterator_index + 1

		if self.iterator_index >= len(self):
			raise StopIteration
		return self[self.iterator_index]

	def __enter__(self):
		"""Called when entering a `with` block"""
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')
		return self

	def __exit__(self, *args):
		"""Called when exiting a `with` block"""
		self.close()

	def __repr__(self):
		if len(self) <= 0:
			return '<TraceSet (0), empty>'
		else:
			return '<TraceSet ({0:d}), {1:s}, ... ,{2:s}>'.format(len(self), repr(self[0]), repr(self[-1]))

	def __len__(self):
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')

		return self.engine.length()

	def __delitem__(self, index):
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')

		if self.engine.is_read_only():
			raise TypeError('Cannot modify trace set, it is (opened) read-only')

		return self.engine.del_traces(index)

	def __setitem__(self, index, traces):
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')

		if self.engine.is_read_only():
			raise TypeError('Cannot modify trace set, it is (opened) read-only')

		# Make sure we have iterable traces
		if isinstance(traces, Trace):
			traces = [traces]

		# Make sure we only are setting traces
		if any(not isinstance(trace, Trace) for trace in traces):
			raise TypeError('All objects assigned to a trace set needs to be of type \'Trace\'')

		return self.engine.set_traces(index, traces)

	def __getitem__(self, index):
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')

		traces = self.engine.get_traces(index)

		# Return the select item(s)
		if isinstance(index, slice):
			return traces
		else:
			# Earlier logic should ensure traces contains one element!
			return traces[0]

	def is_closed(self):
		return self.engine.is_closed()

	def close(self):
		if self.engine is not None:
			self.engine.close()

	def append(self, trace):
		self[len(self):len(self)] = trace

	def extend(self, traces):
		self[len(self):len(self)] = traces

	def insert(self, index, trace):
		# TODO: Not yet implemented nicely
		#self[index:index] = trace
		raise NotImplementedError('Insert has not yet been implemented for TraceSet, please raise a Github ticket for priority!')

	def reverse(self):
		return self[::-1]

	def update_headers(self, headers):
		if self.engine.is_closed():
			raise ValueError('I/O operation on closed trace set')

		return self.engine.update_headers(headers)

	def update_header(self, header, value):
		return self.update_headers({header: value})

	def get_headers(self):
		return self.engine.headers

	def get_header(self, header):
		return self.engine.headers[header]

	def __eq__(self, other):
		"""Compares two trace sets to each other"""
		if not isinstance(other, TrsFile):
			return False

		if len(self) != len(other):
			return False

		# Not using any, because we want to stop as soon as a difference arises
		for self_trace, other_trace in zip(self, other):
			if self_trace != other_trace:
				return False

		return True



import unittest
import trsfile
from os.path import dirname, abspath

class TestGeneric(unittest.TestCase):
	def test_open(self):
		with trsfile.open(dirname(abspath(__file__)) + '/data/90x500xfloat.trs', engine='TrsEngine') as trs_file:
			self.assertIsInstance(trs_file, trsfile.TraceSet)
			self.assertIsInstance(trs_file.engine, trsfile.Engine)
			self.assertIsInstance(trs_file.engine, trsfile.TrsEngine)

			# Assume both handles to be not closed
			self.assertFalse(trs_file.is_closed())

	def test_close(self):
		trs_file = trsfile.open(dirname(abspath(__file__)) + '/data/90x500xfloat.trs', engine='TrsEngine')
		self.assertIsInstance(trs_file, trsfile.TraceSet)
		self.assertIsInstance(trs_file.engine, trsfile.Engine)
		self.assertIsInstance(trs_file.engine, trsfile.TrsEngine)
		trs_file.close()

		# Assume both handles to be closed
		self.assertTrue(trs_file.is_closed())

if __name__ == '__main__':
	unittest.main()



from enum import Enum

class TracePadding(Enum):
	"""Defines the padding mode of the samples in each trace. This can be helpful
	when not all traces will be the same length. This can be set in
	:py:func:`trsfile.open()`, :py:func:`trsfile.trs_open()`

	+----------+---------------------------------------------------------------+
	| Mode     | Description                                                   |
	+==========+===============================================================+
	| NONE     | No padding will be used and an exception will be thrown when  |
	|          | traces are not of the same length.                            |
	+----------+---------------------------------------------------------------+
	| PAD      | All traces will be padded with zeroes to the maximum trace    |
	|          | length.                                                       |
	+----------+---------------------------------------------------------------+
	| TRUNCATE | All traces will be truncated to the minimum trace length.     |
	+----------+---------------------------------------------------------------+
	| AUTO     | Traces will be clipped or padded in the best possible way the |
	|          | storage engine supports. This could mean data is lost which   |
	|          | because retroactive padding is not supported.                 |
	+----------+---------------------------------------------------------------+
	"""

	NONE = 0
	PAD = 1
	TRUNCATE = 2
	AUTO = 3

class SampleCoding(Enum):
	"""Defines the encoding of all the samples in the trace.
	Bit 4 specifies if it is a float (1) or an integer (0),
	bits 0 to 3 specifies the length of the value.
	Finally, bits 5-7 are currently reserved and set to 000.

	This class is just a simple lookup table.
	"""

	BYTE  = (0x01, 1, 'int8')
	SHORT = (0x02, 2, 'int16')
	INT   = (0x04, 4, 'int32')
	FLOAT = (0x14, 4, 'float32')

	def __new__(cls, coding, size, format):
		obj = object.__new__(cls)
		obj._value_ = coding
		obj.size = size
		obj.format = format
		return obj

	@property
	def is_float(self):
		return (self._value_ & 0x10) != 0

class Header(Enum):
	"""All headers that are currently supported in the .trs file format as
	defined in the inspector manual (2018). The storage engine shall try to
	always store all headers regardless if they are used or not. However, some
	file formats will have no way of storing arbitrary headers. As such, optional
	headers can be dropped.

	Some headers can be used by :py:class:`trsfile.trace_set.TraceSet` or
	:py:class:`trsfile.trace.Trace` to augment their functionality. An example
	of this	is the :py:meth:`trsfile.trace.Trace.get_key` method.
	"""
	# Enumeration                       Tag   Name  Mandatory Python type   Length  Default value      Description
	NUMBER_TRACES                    = (0x41, 'NT', True,     int,          4,      0,                 'Number of traces')
	NUMBER_SAMPLES                   = (0x42, 'NS', True,     int,          4,      None,              'Number of samples per trace')
	SAMPLE_CODING                    = (0x43, 'SC', True,     SampleCoding, 1,      SampleCoding.BYTE, 'Sample Coding (see SampleCoding class)')
	LENGTH_DATA                      = (0x44, 'DS', False,    int,          2,      None,              'Length of cryptographic data included in trace')
	TITLE_SPACE                      = (0x45, 'TS', False,    int,          1,      255,               'Title space reserved per trace')
	TRACE_TITLE                      = (0x46, 'GT', False,    str,          None,   'trace',           'Global trace title')
	DESCRIPTION                      = (0x47, 'DC', False,    str,          None,   None,              'Description')
	OFFSET_X                         = (0x48, 'XO', False,    int,          4,      0,                 'Offset in X-axis for trace representation')
	LABEL_X                          = (0x49, 'XL', False,    str,          None,   None,              'Label of X-axis')
	LABEL_Y                          = (0x4A, 'YL', False,    str,          None,   None,              'Label of Y-axis')
	SCALE_X                          = (0x4B, 'XS', False,    float,        4,      1,                 'Scale value for X-axis')
	SCALE_Y                          = (0x4C, 'YS', False,    float,        4,      1,                 'Scale value for Y-axis')
	TRACE_OFFSET                     = (0x4D, 'TO', False,    int,          4,      0,                 'Trace offset for displaying trace numbers')
	LOGARITHMIC_SCALE                = (0x4E, 'LS', False,    int,          1,      0,                 'Logarithmic scale')
	ACQUISITION_RANGE_OF_SCOPE       = (0x55, 'RG', False,    float,        4,      0,                 'Range of the scope used to perform acquisition')
	ACQUISITION_COUPLING_OF_SCOPE    = (0x56, 'CL', False,    int,          4,      0,                 'Coupling of the scope used to perform acquisition')
	ACQUISITION_OFFSET_OF_SCOPE      = (0x57, 'OS', False,    float,        4,      0,                 'Offset of the scope used to perform acquisition')
	ACQUISITION_INPUT_IMPEDANCE      = (0x58, 'II', False,    float,        4,      0,                 'Input impedance of the scope used to perform acquisition')
	ACQUISITION_DEVICE_ID            = (0x59, 'AI', False,    bytes,        None,   None,              'Device ID of the scope used to perform acquisition')
	ACQUISITION_TYPE_FILTER          = (0x5A, 'FT', False,    int,          4,      0,                 'The type of filter used during acquisition')
	ACQUISITION_FREQUENCY_FILTER     = (0x5B, 'FF', False,    float,        4,      0,                 'Frequency of the filter used during acquisition')
	ACQUISITION_RANGE_FILTER         = (0x5C, 'FR', False,    float,        4,      0,                 'Range of the filter used during acquisition')
	TRACE_BLOCK                      = (0x5F, 'TB', True,     None,         0,      None,              'Trace block marker: an empty TLV that marks the end of the header')
	EXTERNAL_CLOCK_USED              = (0x60, 'EU', False,    bool,         1,      False,             'External clock used')
	EXTERNAL_CLOCK_THRESHOLD         = (0x61, 'ET', False,    float,        4,      0,                 'External clock threshold')
	EXTERNAL_CLOCK_MULTIPLIER        = (0x62, 'EM', False,    int,          4,      0,                 'External clock multiplier')
	EXTERNAL_CLOCK_PHASE_SHIFT       = (0x63, 'EP', False,    int,          4,      0,                 'External clock phase shift')
	EXTERNAL_CLOCK_RESAMPLER_MASK    = (0x64, 'ER', False,    int,          4,      0,                 'External clock resampler mask')
	EXTERNAL_CLOCK_RESAMPLER_ENABLED = (0x65, 'RE', False,    bool,         1,      False,             'External clock resampler enabled')
	EXTERNAL_CLOCK_FREQUENCY         = (0x66, 'EF', False,    float,        4,      0,                 'External clock frequency')
	EXTERNAL_CLOCK_BASE              = (0x67, 'EB', False,    int,          4,      0,                 'External clock time base')
	NUMBER_VIEW                      = (0x68, 'VT', False,    int,          4,      0,                 'View number of traces: number of traces to show on opening')
	TRACE_OVERLAP                    = (0x69, 'OV', False,    bool,         1,      False,             'Overlap: whether to overlap traces in case of multi trace view')
	GO_LAST_TRACE                    = (0x6A, 'GL', False,    bool,         1,      False,             'Go to last trace on opening')
	INPUT_OFFSET                     = (0x6B, 'IO', False,    int,          4,      0,                 'Input data offset in trace data')
	OUTPUT_OFFSET                    = (0x6C, 'OO', False,    int,          4,      0,                 'Output data offset in trace data')
	KEY_OFFSET                       = (0x6D, 'KO', False,    int,          4,      0,                 'Key data offset in trace data')
	INPUT_LENGTH                     = (0x6E, 'IL', False,    int,          4,      0,                 'Input data length in trace data')
	OUTPUT_LENGTH                    = (0x6F, 'OL', False,    int,          4,      0,                 'Output data length in trace data')
	KEY_LENGTH                       = (0x70, 'KL', False,    int,          4,      0,                 'Key data length in trace data')

	def __new__(cls, tag, tag_name, is_mandatory, type, length, default, description):
		obj = object.__new__(cls)
		obj._value_ = tag
		obj.tag_name = tag_name
		obj.is_mandatory = is_mandatory
		obj.type = type
		obj.length = length
		obj.default = default
		obj.description = description
		return obj

	@classmethod
	def has_value(cls, tag):
		return any(tag == item.value for item in cls)

	@classmethod
	def get_mandatory(cls):
		return set([header for header in cls if header.is_mandatory])



import unittest
import trsfile
import binascii
from os.path import dirname, abspath

class TestTrsFile(unittest.TestCase):
	def setUp(self):
		self.trs_file = trsfile.open(dirname(abspath(__file__)) + '/data/90x500xfloat.trs', engine='TrsEngine')

	def tearDown(self):
		self.trs_file.close()

	def test_iterator(self):
		"""Check if we can iterate over the samples"""
		i = 0
		for samples in self.trs_file[0]:
			i += 1
		self.assertEqual(i, len(self.trs_file[0]))

	def test_data(self):
		self.assertEqual(self.trs_file[0].data, binascii.unhexlify('43B94E34D3A221B27640C5AD87FBE5DF'))

	def test_title(self):
		self.assertEqual(self.trs_file[0].title, 'Clipped trace')

	def test_equality(self):
		self.assertEqual(self.trs_file[0], self.trs_file[0])
		self.assertNotEqual(self.trs_file[0], self.trs_file[1])

if __name__ == '__main__':
	unittest.main()



import shutil
import time
import numpy
import pickle
from pathlib import Path

from trsfile.trace import Trace
from trsfile.common import Header, SampleCoding, TracePadding
from trsfile.engine.engine import Engine

class FileEngine(Engine):
	"""
	This engine tries to save traces to disk in the most versatile and simple
	manner available. No known tools support this file format and serve only as
	an intermediate step to later convert it to a supported format.

	This is can be useful when the trace length (number of samples) varies as
	this is often not supported in trace files.

	After acquisition, the file can be converted to the proper format with the
	correct padding mode.

	This engine supports the following options:

	+--------------+-----------------------------------------------------------+
	| Option       | Description                                               |
	+==============+===========================================================+
	| headers      | Dictionary containing zero or more headers, see           |
	|              | :py:class:`trsfile.common.Header`                         |
	+--------------+-----------------------------------------------------------+
	"""

	INFO_FILE = 'traceset.pickle'

	def __get_trace_path(self, i, name):
		return (self.path / '{0:d}.{1:s}'.format(i, name))

	def __init__(self, path, mode = 'x', **options):
		# Defaults
		self.path = Path(path)
		self.headers = {}
		self.read_only = False

		# Shadow list of traces in files
		self.shadow_trace_index = -1
		self.shadow_traces = []

		# Parse the mode
		if mode == 'r':
			"""r = open for reading"""
			if not self.path.is_dir() or not (self.path / self.INFO_FILE).is_file():
				raise FileNotFoundError('Path \'{0:s}\' does not point to a tmp trace set'.format(path))

			# Load the headers
			with (self.path / self.INFO_FILE).open('rb') as f:
				self.headers = pickle.load(f)

			# Initialize the shadow_traces list
			self.shadow_traces = sorted([int(trace_path.stem) for trace_path in self.path.glob('*.samples')])
			self.shadow_trace_index = max(self.shadow_traces) + 1

			self.read_only = True

		elif mode == 'w':
			"""open for writing, truncating the file first"""

			# Remove the directory if it exists
			if self.path.is_dir():
				shutil.rmtree(str(self.path), True)

				# Wait until it is removed
				try:
					while self.path.is_dir():
						time.sleep(0.001)
				except:
					pass

			# Create the temporary folder and initialize this class
			self.path.mkdir()
			self.__initialize_headers()

		elif mode == 'x':
			"""open for exclusive creation, failing if the file already exists (default)"""
			if self.path.is_dir():
				raise FileExistsError('Trace set already exists at path \'{0:s}\''.format(str(self.path)))

			# Create the temporary folder and initialize this class
			self.path.mkdir()
			self.__initialize_headers()

		elif mode == 'a':
			"""a = open for writing, appending to the end of the file if it exists"""
			if self.path.is_dir() and (self.path / self.INFO_FILE).is_file():
				# Load the headers
				with (self.path / self.INFO_FILE).open('rb') as f:
					self.headers = pickle.load(f)

				# Initialize the shadow_traces list
				self.shadow_traces = sorted([int(trace_path.stem) for trace_path in self.path.glob('*.samples')])
				self.shadow_trace_index = max(self.shadow_traces) + 1
			else:
				# Create the temporary folder and initialize this class
				self.path.mkdir()
				self.__initialize_headers()

		else:
			raise ValueError('invalid mode: \'{0:s}\''.format(mode))

		# Update the headers
		headers = options.get('headers', None)
		if self.is_read_only() and headers is not None:
			raise ValueError('Cannot add headers when opening in read-only mode')
		elif headers is not None:
			self.update_headers(headers)

	def __initialize_headers(self):
		headers = {}

		# Let's support dynamic sample coding depending on the trace
		headers[Header.SAMPLE_CODING] = None

		# Add any mandatory headers that are missing
		for header in Header.get_mandatory():
			if not header in headers:
				headers[header] = header.default

		# Store these default headers
		self.update_headers(headers)

	def update_headers(self, headers):
		changed_headers = super().update_headers(headers)
		if len(changed_headers) > 0:
			# Dump all headers to disk
			with (self.path / self.INFO_FILE).open('wb') as f:
				pickle.dump(self.headers, f)

	def is_closed(self):
		return not self.path.is_dir() or not (self.path / self.INFO_FILE).is_file()

	def length(self):
		return len(self.shadow_traces)

	def del_traces(self, index):
		# Remove the shadow traces and with that check if indexes are correct
		exception = None
		try:
			indices = self.shadow_traces[index]
			del self.shadow_traces[index]

			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as err:
			exception = err

		# Do we have an exception, re-raise
		if exception is not None:
			raise IndexError(exception)

		# Delete all traces on the file system
		for trace_index in indices:
			for category in ['title', 'data', 'samples']:
				path = self.__get_trace_path(trace_index, category)
				if path.is_file():
					path.unlink()

	def get_traces(self, index):
		# Try access, and re-raise if wrong for fancy indexing errors
		try:
			indices = self.shadow_traces[index]
			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as exception:
			raise IndexError(exception)

		# Now obtain all requested traces from file
		traces = []
		for i in indices:
			# Read the samples
			path = self.__get_trace_path(i, 'samples')
			if path.is_file():
				with path.open('rb') as tmp_file:
					# First byte is always sample coding
					sample_coding = SampleCoding(tmp_file.read(1)[0])
					samples = numpy.fromfile(tmp_file, sample_coding.format, -1)
			else:
				raise IOError('Unable to read samples from trace {0:d}'.format(i))

			# Title
			path = self.__get_trace_path(i, 'title')
			if path.is_file():
				with path.open('rb') as tmp_file:
					title = tmp_file.read().decode('utf-8')
			else:
				title = Header.TRACE_TITLE.default

			# Read the data
			path = self.__get_trace_path(i, 'data')
			if path.is_file():
				with path.open('rb') as tmp_file:
					data = tmp_file.read()
			else:
				data = b''

			# Create trace and make sure headers point to our dict
			traces.append(Trace(sample_coding, samples, data, title, self.headers))

		return traces

	def set_traces(self, index, traces):
		# Make sure we have iterable traces
		if isinstance(traces, Trace):
			traces = [traces]

		# Get all traces that we are going to remove
		try:
			indices = self.shadow_traces[index]
			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as exception:
			raise IndexError(exception)

		# Remove the traces from disk only to keep storage lean and mean
		for trace_index in indices:
			for category in ['title', 'data', 'samples']:
				path = self.__get_trace_path(trace_index, category)
				if path.is_file():
					path.unlink()

		# Store all traces with the next sequence numbers and keep these numbers as a list
		new_traces = []
		for trace in traces:
			self.shadow_trace_index += 1
			new_traces.append(self.shadow_trace_index)

			# Save the trace data
			# Write the title as ascii
			with self.__get_trace_path(self.shadow_trace_index, 'title').open('wb') as tmp_file:
				tmp_file.write(trace.title if not isinstance(trace.title, str) else trace.title.encode('utf-8'))

			# Write the data file
			if trace.data is not None and len(trace.data) > 0:
				with open(self.__get_trace_path(self.shadow_trace_index, 'data'), 'wb') as tmp_file:
					tmp_file.write(trace.data)

			# Write the sample file
			with open(self.__get_trace_path(self.shadow_trace_index, 'samples'), 'wb') as tmp_file:
				tmp_file.write(bytes([trace.sample_coding.value]))
				trace.samples.tofile(tmp_file)

		# Now we just assign the new_traces however, the slicing works
		if isinstance(index, slice):
			self.shadow_traces[index] = new_traces
		else:
			if len(new_traces) != 1:
				raise TypeError('assigning multiple new traces to single trace')
			self.shadow_traces[index] = new_traces[0]

	def close(self):
		# We do not need to close anything :)
		pass



import os
import trsfile
import time
import tempfile
import unittest
import math
import shutil
from trsfile import Trace, SampleCoding, Header, TracePadding

def get_sample(x):
	return math.sin(0.5 * x) * 1000

class TestCreation(unittest.TestCase):

	def setUp(self):
		# We want a tmp file name in a safe way
		fd, self.tmp_path = tempfile.mkstemp(prefix='trsfile_')

		# However, some tests require the file to not yet exist
		os.close(fd)
		os.remove(self.tmp_path)

		# Small guard for windows to prevent some weird race conditions...
		time.sleep(0.01)

	def tearDown(self):
		# Perform any cleanup of the file
		try:
			if os.path.isfile(self.tmp_path):
				os.remove(self.tmp_path)
			elif os.path.isdir(self.tmp_path):
				shutil.rmtree(self.tmp_path, True)
				time.sleep(0.01)
		except FileNotFoundError:
			pass

	def test_write(self):
		trace_count = 100
		sample_count = 1000

		try:
			with trsfile.open(self.tmp_path, 'w', headers = {
					Header.LABEL_X: 'Testing X',
					Header.LABEL_Y: 'Testing Y',
					Header.DESCRIPTION: 'Testing trace creation',
				}, padding_mode = TracePadding.AUTO) as trs_traces:
				trs_traces.extend([
					Trace(
						SampleCoding.FLOAT,
						[0] * sample_count,
						data = i.to_bytes(8, byteorder='big')
					)
					for i in range(0, trace_count)]
				)
		except Exception:
			self.assertTrue(False)

	def test_write_closed(self):
		trace_count = 100
		sample_count = 1000

		with trsfile.open(self.tmp_path, 'w', padding_mode=TracePadding.AUTO) as trs_traces:
			trs_traces.extend([
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)]
			)

		# Should raise an "ValueError: I/O operation on closed trace set"
		with self.assertRaises(ValueError):
			print(trs_traces)

	def test_read(self):
		trace_count = 100
		sample_count = 1000

		original_traces = [
				Trace(
					SampleCoding.FLOAT,
					[get_sample(i) for i in range(0, sample_count)],
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)
			]

		# Create a trace
		with trsfile.open(self.tmp_path, 'w', headers = {
				Header.LABEL_X: 'Testing X',
				Header.LABEL_Y: 'Testing Y',
				Header.DESCRIPTION: 'Testing trace creation',
			}, padding_mode = TracePadding.AUTO
			) as trs_traces:

			trs_traces.extend(original_traces)

			# Make sure length is equal
			self.assertEqual(len(original_traces), len(trs_traces))

		# Read the trace and check if everything is good
		with trsfile.open(self.tmp_path, 'r') as trs_traces:
			# Check if lengths are still good :)
			self.assertEqual(len(original_traces), len(trs_traces))

			# Check if every trace is saved correctly
			for original_trace, trs_trace in zip(trs_traces, original_traces):
				self.assertEqual(original_trace, trs_trace)

	def test_read_non_existing(self):
		with self.assertRaises(FileNotFoundError):
			with trsfile.open(self.tmp_path, 'r') as trs_traces:
				pass

	def test_append(self):
		trace_count = 100
		sample_count = 1000

		# Append to a non-existing file, behaves same as normal "write"
		with trsfile.open(self.tmp_path, 'a', padding_mode=TracePadding.AUTO) as trs_traces:
			self.assertEqual(len(trs_traces), 0)

			# Extend the trace file with 100 traces with each 1000 samples
			trs_traces.extend([
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)]
			)

			self.assertEqual(len(trs_traces), trace_count)

		# Now open and close for a few times while adding some number of traces
		expected_length = trace_count
		for t in range(0, 10):

			trace_count = (t + 1) * 10

			with trsfile.open(self.tmp_path, 'a', padding_mode=TracePadding.AUTO) as trs_traces:
				self.assertEqual(len(trs_traces), expected_length)

				# Extend the trace file with 100 traces with each 1000 samples
				trs_traces.extend([
					Trace(
						SampleCoding.FLOAT,
						[0] * sample_count,
						data = i.to_bytes(8, byteorder='big')
					)
					for i in range(0, trace_count)]
				)

				expected_length += trace_count
				self.assertEqual(len(trs_traces), expected_length)

	def test_exclusive(self):
		trace_count = 100
		sample_count = 1000

		# Write to file exclusively
		with trsfile.open(self.tmp_path, 'x', padding_mode=TracePadding.AUTO) as trs_traces:
			self.assertEqual(len(trs_traces), 0)

			# Extend the trace file with 100 traces with each 1000 samples
			trs_traces.extend([
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)]
			)

			self.assertEqual(len(trs_traces), trace_count)

		# Now try again (this should throw an exception)
		with self.assertRaises(FileExistsError):
			with trsfile.open(self.tmp_path, 'x') as trs_traces:
				self.assertEqual(len(trs_traces), trace_count)


	def test_extend(self):
		trace_count = 100
		sample_count = 1000

		with trsfile.open(self.tmp_path, 'w', padding_mode=TracePadding.AUTO) as trs_traces:
			# Extend empty list
			trs_traces.extend([])
			self.assertEqual(len(trs_traces), 0)

			# Extend non empty list
			trs_traces.extend([
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = b'\x00' * 8
				)
				]
			)
			self.assertEqual(len(trs_traces), 1)

			# Extend non empty list
			trs_traces.extend([
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)]
			)
			self.assertEqual(len(trs_traces), trace_count + 1)

	def test_padding_none(self):
		sample_count = 1000

		with trsfile.open(
			self.tmp_path,
			'w',
			padding_mode = TracePadding.NONE,
			headers = {
				Header.NUMBER_SAMPLES: sample_count,
				Header.LENGTH_DATA: 8,
				Header.SAMPLE_CODING: SampleCoding.FLOAT
			}
			) as trs_traces:
			# This is the length of the trace
			trs_traces.extend(
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = b'\x00' * 8
				)
			)

			# Length is smaller
			with self.assertRaises(ValueError):
				trs_traces.extend(
					Trace(
						SampleCoding.FLOAT,
						[0] * (sample_count - 1),
						data = b'\x10' * 8
					)
				)
			self.assertEqual(len(trs_traces), 1)

			# Length is bigger
			with self.assertRaises(ValueError):
				trs_traces.extend(
					Trace(
						SampleCoding.FLOAT,
						[0] * (sample_count + 1),
						data = b'\x01' * 8
					)
				)
			self.assertEqual(len(trs_traces), 1)

			# Length is equal
			trs_traces.extend(
				Trace(
					SampleCoding.FLOAT,
					[0] * sample_count,
					data = b'\x00' * 8
				)
			)
			self.assertEqual(len(trs_traces), 2)

	def test_padding(self):
		trace_count = 100
		sample_count = 1000
		fmt = SampleCoding.FLOAT

		with trsfile.open(self.tmp_path, 'w', padding_mode=TracePadding.AUTO) as trs_traces:
			# This is the length everything should be padded/clipped to
			trs_traces.extend(
				Trace(
					fmt,
					b'\xDE' * (sample_count * fmt.size),
					data = b'\x00' * 8
				)
			)

			# Padding mode
			trs_traces.extend([
				Trace(
					fmt,
					b'\xDE' * (sample_count + i) * fmt.size,
					data = abs(i).to_bytes(8, byteorder='big')
				)
				for i in range(0, -trace_count, -1)]
			)

			# Clipping mode
			trs_traces.extend([
				Trace(
					fmt,
					b'\xDE' * (sample_count + i) * fmt.size,
					data = i.to_bytes(8, byteorder='big')
				)
				for i in range(0, trace_count)]
			)

		with trsfile.open(self.tmp_path, 'r') as trs_traces:
			self.assertEqual(len(trs_traces), trace_count * 2 + 1)

			# Check that all traces are of the same size
			for trs_trace in trs_traces:
				self.assertEqual(len(trs_trace), sample_count)

			# Check that all padding is zero
			for i, trs_trace in enumerate(trs_traces[1:101]):
				# Difficult case :)
				if i == 0:
					continue

				for si, sample in enumerate(trs_trace[-i:]):
					self.assertEqual(sample, 0.0 if fmt == SampleCoding.FLOAT else 0, str(i))

				# Test that this is indeed not zero
				self.assertNotEqual(trs_trace[-i - 1], 0)

if __name__ == '__main__':
	unittest.main()




from trsfile.common import Header

class Engine:
	read_only = False

	def __init__(self, path, mode = 'x', **options):
		"""Initializes a storage engine for the specified format.

		:param path: relative or absolute path to the file
		:param mode: in which mode to open the file (r, w, a, x)
		:param **options: options that this engine supports, the Engine specifies
			which options are supported.
		:type path: str
		:type mode: str
		:type **options: dict(str, any)
		"""
		raise NotImplementedError('Implement \'__init__\' for the storage engine')

	def close(self):
		"""Closes the file backing the trace set. It also could perform the
		final writes to synchronize the file with the trace set.

		:returns: None
		"""
		raise NotImplementedError('Implement \'close\' for the storage engine')

	def length(self):
		"""Returns the total number of traces

		:returns: total number of traces
		:rtype: int
		"""
		raise NotImplementedError('Implement \'length\' for the storage engine')

	def is_closed(self):
		"""Returns if the file backing the trace set is closed

		:returns: True if the file is closed, otherwise False
		:rtype: boolean
		"""
		raise NotImplementedError('Implement \'is_closed\' for the storage engine')

	def is_read_only(self):
		"""Returns if the trace set is read-only

		:returns: True if the file is read-only, otherwise False
		:rtype: boolean
		"""
		return self.read_only

	# Functions that are optionally implemented
	def del_traces(self, index):
		"""Deletes zero or more traces from the trace set

		:param index: the slice or index that specifies which traces to delete
		:type index: slice, int
		:returns: None
		"""
		raise TypeError('Cannot remove traces with this storage engine')

	def set_traces(self, index, traces):
		"""Inserts zero or more traces into the trace set

		:param index: the slice or index that specifies were to insert traces
		:param traces: zero or more traces to insert into the trace set
		:type index: slice, int
		:type traces: Trace, list[Trace]
		:returns: None
		"""
		raise TypeError('Cannot change traces with this storage engine')

	def get_traces(self, index):
		"""Retrieves zero or more traces from the trace set

		:param index: the slice or index that specifies which traces to get
		:type index: slice, int
		:returns: a list of zero or more traces from the trace set
		:rtype: Trace, list[Trace]
		"""
		raise TypeError('Cannot get traces with this storage engine')

	# The headers as defined by the TRS specifications
	def update_headers(self, headers):
		"""Updates zero or more headers

		:param headers: dictionary of header, value pairs to update
		:type headers: dict(Header, any)
		:returns: a list of the headers that changed
		:rtype: list[Header]
		"""
		# Early return
		if headers is None:
			return []

		if self.is_read_only():
			raise TypeError('Cannot modify trace set, it is (opened) read-only')

		if not isinstance(headers, dict) or any([not isinstance(header, Header) for header in self.headers]):
			raise TypeError('All headers have to be of type \'Header\'')

		# TODO: We can test the header type here, do we want to?

		# Only update headers that are changed
		changed_headers = {}
		for header, value in headers.items():
			if header in self.headers and value == self.headers[header]:
				continue
			changed_headers[header] = value

		# Do nothing if nothing has changed
		if len(changed_headers) <= 0:
			return []

		# Update internally the headers
		self.headers.update(changed_headers)

		# Notify that headers have been changed
		return changed_headers

	def update_header(self, header, value):
		"""Updates one specific header

		:param header: header to update
		:param value: value of the header to update
		:type header: Header
		:type value: any
		:returns: a list of the headers that changed
		:rtype: list[Header]
		"""
		return self.update_headers({header: value})



"""
The defactor trace set (.trs files) library for reading
Riscure Inspector trace files.
"""

name        = "trsfile"
__version__ = '0.3.0'
__author__  = 'Kevin Valk'
__contact__ =  'valk@riscure.com'
__all__     = [
'trs_open', 'Trace', 'TraceSet', 'SampleCoding', 'Header', 'TracePadding', \
'Engine', 'TrsEngine', 'FileEngine',
]

from trsfile.trace import Trace
from trsfile.common import Header, SampleCoding, TracePadding
from trsfile.trace_set import TraceSet
from trsfile.engine.engine import Engine

# All our engines
from trsfile.engine.trs import TrsEngine
from trsfile.engine.file import FileEngine

def trs_open(path, mode = 'r', **options):
	"""Reads, modifies or creates a :py:obj:`TraceSet` with a specific storage
	engine (defaults to :py:obj:`TrsEngine`).

	:param path: path to the file or directory
	:param mode: mode how to open the file or directory (same as the default
		Python :py:obj:`open`)
	:param options: zero or more options that are passed down to the
		:py:obj:`TraceSet` and the storage engine. Available options can be found
		in the different storage engines. The storage engine can be selected with
		:code:`engine = 'TrsEngine'` (default value).
	:type path: str
	:type mode: str
	:type options: dict(str, any)
	:returns: instance of a new or initialized :py:obj:`TraceSet`
	:rtype: TraceSet
	"""
	return TraceSet(path, mode, **options)

open = trs_open
